/**
   The static method parse() is a recursive descent
   parser for strings that hold the infix notation defined
   by the following BNF grammar.

         Expr -> Expr ( '+' | '-' ) Term | Term

         Term -> Term ( '*' | '-' ) Signed | Signed

       Signed -> '-' Factor | Factor

       Factor -> Base '^' Signed | Base

        Base  -> '(' Expr ')' | 'sqrt' Base | Number

       Number -> any valid Java double

   In a recursive descent parser, every production is handled
   by a method. Since this grammar has six productions, the
   parser consists of six methods, getExpr(), getTerm(),
   getSigned(), getFactor(), getBase(), and getNumber().

   But this grammar is left recursive, so we cannot use it
   to write a recursive descent parser. We need to rewrite
   the grammar in EBNF, and "factor out" the recursion.

         Expr ->   Term ( ('+'|'-') Term )*

         Term -> Signed ( ('*'|'/') Signed )*

       Signed -> [ '-' ] Factor

       Factor -> Base [ '^' Signed ]

       Base -> '(' Expr ')' | 'sqrt' Base | Number

       Number -> any valid Java double

   The Expr and Term productions are now no longer recursive
   (but they are still mutually recursive). The recursion has
   been replaced with "iteration". The getExpr() method should
   use a while-loop to determine how many '+|-' operators it needs
   to parse. And the getTerm() method should also use a while-loop
   to determine how many '*|/' operators it needs to parse.

   Input strings are "tokenized" by the very bare bones
   tokenizer contained in the file Tokenizer.java. This
   tokenizer requires that all tokens in the input string
   be surrounded by white space.
*/

public class Parse
{
   public static boolean parse(String expr) throws ParseException
   {
      Tokenizer tokens = new Tokenizer(expr);

      boolean result = getExpr(tokens); // parse the token stream

      if ( tokens.hasToken() )  // there shouldn't be any more tokens
         throw new ParseException("syntax error: unexpected input: "+tokens.peekToken()+"\n"+tokens);

      return result;
   }//parse()



   private static boolean getExpr(Tokenizer tokens) throws ParseException
   {
      boolean result = getTerm( tokens );

      // check for operators at this precedence level
      while ( tokens.hasToken() && "+-".indexOf(tokens.peekToken()) != -1 )
      {
         String op = tokens.nextToken(); // consume the operator token

         result = getTerm(tokens);
      }
      return result;
   }//getExpr()


   private static boolean getTerm(Tokenizer tokens) throws ParseException
   {
      boolean result = getSigned( tokens );

      // check for operators at this precedence level
      while ( tokens.hasToken() && "*/".indexOf(tokens.peekToken()) != -1 )
      {
         String op = tokens.nextToken(); // consume the operator token

         result = getSigned(tokens);
      }
      return result;
   }//getTerm()


   private static boolean getSigned(Tokenizer tokens) throws ParseException
   {
      boolean result = false;

      if ( tokens.hasToken() && tokens.peekToken().equals("-") )
      {
         String op = tokens.nextToken(); // consume the operator token

         result = getFactor(tokens);
      }
      else
      {
         result = getFactor(tokens);
      }
      return result;
   }//getSigned()


   private static boolean getFactor(Tokenizer tokens) throws ParseException
   {
      boolean result = getBase( tokens );

      if ( tokens.hasToken() && tokens.peekToken().equals("^") )
      {
         String op = tokens.nextToken(); // consume the operator token

         result = getSigned(tokens);
      }
      return result;
   }//getFactor()


   private static boolean getBase(Tokenizer tokens) throws ParseException
   {
      boolean result = false;

      if ( ! tokens.hasToken() )
      {
         throw new ParseException("syntax error: unexpected end of input:\n"+tokens);
      }
      else if ( tokens.peekToken().equals("(") )
      {
         String notNeeded = tokens.nextToken(); // consume the "("

         result = getExpr(tokens);

         if ( ! tokens.hasToken() )
         {
            throw new ParseException("syntax error: unexpected end of input:\n"+tokens);
         }
         if (! tokens.match(")") ) // consume the ")"
         {
            throw new ParseException("syntax error: expected \")\", found : "+tokens.peekToken());
         }
      }
      else if( tokens.peekToken().equals("sqrt") )
      {
         String op = tokens.nextToken(); // consume the operator token

         result = getBase(tokens);
      }
      else  // must be a number
      {
         result = getNumber(tokens);
      }
      return result;
   }//getBase()


   private static boolean getNumber(Tokenizer tokens) throws ParseException
   {
      if ( ! tokens.hasToken() )
      {
         throw new ParseException("syntax error: unexpected end of input:\n"+tokens);
      }

      // a number is a single token
      String tk = tokens.nextToken();
      try
      {
         Double d = Double.parseDouble(tk);
      }
      catch (NumberFormatException nfe)
      {
         throw new ParseException("syntax error: illegal number: "+tk+"\n"+tokens);
      }
      return true;
   }//getNumber()

}//Parse
